% XFFT - a Matlab toolbox to compute eXtended Fast Fourier transforms
% 
% Copyright (C) 2015 Ines Melzer
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.
% 
% $LastChangedDate: 2015-10-21 13:41:37 +0200 (Mi, 21. Okt 2015) $
% $Author: melzer $
% $Revision: 6165 $
%
%
% Class for fast approximating computing trigonometric sums via the butterfly sparse Fourier transform in the two dimensional case. You can choose by option different types of basis for the local approximations. 
% sparse_FFT2D properties:
%    N       - domain parameter, must be a power of 2
%    MX      - sampling nodes in  X=[0,N]^2 as M1x2 matrix 
%    MOmega  - sampling nodes in Omega=[0,N]^2 as M2x2 matrix
%    p       - approximation rank
%    option  - option defines the basis for the local approximations, possible options are:
%                Mtb: monomial-type basis
%                Ltb: Lagrante-type basis
%                Ltb*: Lagrange-type basis with precomputation of the trigonometric Lagrange-polynomials in the nodes MX
%    invG    - computation of the inverse matrix G^{-1} (option='Mtb')
%    H       - computation of the matrix H (option='Mtb')
%    Lleft   - computation of the Lagrange matrix Lleft for a left son box (option='Ltb','Ltb*')
%    Lright  - computation of the Lagrange matrix Lleft for a right son box (option='Ltb','Ltb*')
%    Laglast - computation of the function values of Lagrange polynomials in MX (option='Ltb*')
%
% sparse_FFT2D methods:
%   sparse_FFT2D  - constructor
%   mtimes        - computing the butterfly sparse Fourier transform 
%
% Examples
%   see Matlab script start2D.m

classdef sparse_FFT2D


properties(SetAccess='private')
    N=[]; %domain parameter
    MX=[];%sampling nodes in the spatial domain
    MOmega=[];%sampling nodes in the frequency domain
    p=[];%approximation rank
    
    option=[];
     
    %matrices for option Mtb
    invG=[]; %matrix G^{-1}
    H=[]; %matrix H
    
    %matrices for option Ltb,Ltb*
    Lleft=[];%Lagrange matrix for a left son
    Lright=[];%Lagrange matrix for a right son
    Laglast=[]; %function values of lagrange polynomials in MX
     
end %properties

properties(Hidden=true,SetAccess='private',GetAccess='private')
  alpha=[]; %grid  in [-1/2,1/2] for A
  beta=[];  %grid  in [-1/2,1/2] for B

  
  %all variables, which are selecting from the trees
  LastlevelA={}; %cellaray with rows {m1A,m2A,indices,(number of sons)}
  LastlevelB={}; %cellaray with rows {m1B,m2B,indices,number of sons}
  ABoxes={}; %2-column matrix [m1A,m2A]
  BBoxes={}; %3-column matrix [m1B,m2B,number of sons]
  numSonsLevel=[]; %number of sons for each level L-l of T_Omega as vector
  
end%properties


methods
  
function plan=sparse_FFT2D(MX,MOmega,p,N,option,varargin)
% Constructor
%
% plan=sparse_FFT2D(MX,MOmega,p,N,option)
% plan=sparse_FFT2D(MX,MOmega,p,N,option,'no_copinfo') % disable copyright info
%
% INPUT
%   for the input arguments see properties of class sparse_FFT2D with the same name
%
% OUTPUT
%   plan  object of type sparse_FFT2D

	nopt=size(varargin,2);
	if( ~(nopt==1 && strcmp(varargin{1},'no_copinfo')))
		sparse_FFT1D.show_copyright;
	end %if

    plan.N=N;
    plan.MX=MX;
    plan.MOmega=MOmega;
    plan.p=p;
    plan.option=option;
    
    %tree T_x
    M1=size(MX,1);
    ind1=1:1:M1;
    TX=tree2D(MX,ind1,plan.N,0,0,0); 
    
    %tree T_Omega
    M2=size(MOmega,1);
    ind2=1:1:M2;
    TOmega=tree2D(MOmega,ind2,plan.N,0,0,0); 
  

    %grids in [-1/2,1/2], here you can choose arbitrary grids in [-1/2,1/2]:
	j=0:1:plan.p-1; 
	plan.alpha=1/2*cos( (2*j+1)./(2*plan.p)*pi); %chebyshev grid first kind
	plan.beta=1/2-j./(p-1); %equispaced
	
    L=log2(N);
    %initialisation with tree functions
    plan.LastlevelA=getBoxes(TX,L);
    plan.LastlevelB=getBoxes(TOmega,L);
    plan.numSonsLevel=zeros(L,1);
    for l=1:L
      plan.ABoxes{l}=getLevelParaA(TX,l);
      plan.BBoxes{l}=getLevelParaB(TOmega,L-l);
      plan.numSonsLevel(l)=countSonsInLevel(TOmega,L-l); %number of sons in level L-l
    end
    
    switch(plan.option)
        case 'Mtb'
            %initialisation of the matrices
            G=zeros(p,p);
            plan.H=zeros(p,p);
            for tmp=1:plan.p
                G(tmp,:)=exp(2*pi*1i*(plan.alpha(tmp)).*(plan.beta.'));
                plan.H(tmp,:)=exp(pi*1i*(plan.alpha(tmp)).*(plan.beta.'));
            end
            plan.invG=inv(G);
        
        case {'Ltb','Ltb*'}
            %Lagrange matrices
            plan.Lleft=zeros(plan.p,plan.p);
            plan.Lright=zeros(plan.p,plan.p);
            for s=1:p
                plan.Lleft(:,s)=lagpol_alpha(0,p,plan.alpha,s);
                plan.Lright(:,s)=lagpol_alpha(1,p,plan.alpha,s);
            end%for s
        
            switch(plan.option)
                case 'Ltb*'
                    display('Ltb with precomputation of the lagrange polynomials in the nodes MX')
                        %additional precomputation of the lagrange polynomials in the nodes MX
                        %evaluation of the lagrange polynomials in the nodes MX 
                        Last_ABoxes=plan.LastlevelA;
                        %wB=plan.N;
                        wA=1;
                        shiftMX=zeros(M1,2);
                        for r=1:size(Last_ABoxes,1)
                            tmpA=Last_ABoxes(r,:);
                            m1A=tmpA{1};
                            m2A=tmpA{2};
                            A_index=tmpA{3};
                            cA=wA*[m1A+1/2,m2A+1/2];
                            Mat_cA=repmat(cA,length(A_index),1);
                            shiftMX(A_index,:)=MX(A_index,:)-Mat_cA;
                        end
						
                        zalpha=exp(-2*pi*1i*plan.alpha/(plan.p-1));
                        shiftMZ=exp(-2*pi*1i*shiftMX/(plan.p-1));
                        tmpLag1=zeros(M1,plan.p);
                        tmpLag2=zeros(M1,plan.p);
                        for s=1:plan.p
                            tmpLag1(:,s)=lagpol_z(plan.p,shiftMZ(:,1),zalpha,s);
                            tmpLag2(:,s)=lagpol_z(plan.p,shiftMZ(:,2),zalpha,s);
                        end
			
                        plan.Laglast=[tmpLag1;tmpLag2]; %save matrices among each other
                otherwise
                    display('Ltb without precomputation of the lagrange polynomials in the nodes MX')
            end%switch
			
           otherwise
			display('unknown option')
    end%switch
end%constructor

end%methods

end%classdef
